Model subclass: #TransportInterface
	instanceVariableNames: 'socket connection stream bossWrite bossRead readProcess writeProcess writeQueue '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smalltalk-Interface'!
TransportInterface comment:
'I represent an interface between two Smalltalk images which is socket based. Two processes
are associated with each interface in order to allow for asynchronous communications. We 
need two BOSS streams as the sender and receiver may be different releases of Smalltalk and
may not have identical formats.

Instance variable		Meaning
==============		=======
socket					<UnixSocketAccessor>	the raw UNIX socket.
connection				<ExternalConnection>	a connection.
stream					<ReadAppendStream>	a byte-oriented stream.
bossRead				<BinaryObjectStorage>	the read boss stream.
bossWrite				<BinaryObjectStorage>	the write boss stream.
readProcess				<Process>				a process handling asynchronous communications.
writeProcess			<Process>				a process handling the physical writes on the socket.
writeQueue				<SharedQueue>			a queue of objects to be written to the socket.

Written by: Tony White (arpw@bnr.ca) and Dwight Deugo (ddeugo@bnr.ca)

This source code is available to all with no restrictions. We only ask that  
proper credit be passed on so that we might hear from those who have found  
the TransportInterface useful.
'!


!TransportInterface methodsFor: 'accessing'!

bossRead
	"This returns the instance variable"

	^bossRead isNil
		ifTrue: [self bossReadInit: stream]
		ifFalse: [bossRead]!

bossRead: aBOSS 
	"This sets the instance variable"

	^bossRead := aBOSS!

bossWrite
	"This returns the instance variable"

	^bossWrite isNil
		ifTrue: [self bossWriteInit: stream]
		ifFalse: [bossWrite]!

bossWrite: aBOSS 
	"This sets the instance variable"

	^bossWrite := aBOSS!

connection
	"This returns the instance variable"

	^connection!

connection: aConnection 
	"This sets the instance variable"

	^connection := aConnection!

readProcess
	"This returns the instance variable"

	^readProcess!

readProcess: aProcess 
	"This sets the instance variable"

	^readProcess := aProcess!

socket
	"This returns the instance variable"

	^socket!

socket: aSocket 
	"This sets the instance variable"

	^socket := aSocket!

stream
	"This returns the instance variable"

	^stream!

stream: aStream 
	"This sets the instance variable"

	^stream := aStream!

writeProcess
	"This returns the instance variable"

	^writeProcess!

writeProcess: aProcess 
	"This sets the instance variable"

	^writeProcess := aProcess!

writeQueue
	"This returns the instance variable"

	^writeQueue!

writeQueue: aSharedQueue 
	"This sets the instance variable"

	^writeQueue := aSharedQueue! !

!TransportInterface methodsFor: 'initialize-release'!

close
	"Close the interface"

	self closeRead; closeWrite; closeStream; closeSocket! !

!TransportInterface methodsFor: 'message processing'!

readProcessLoop
	"Sit in this loop until the socket is closed. We delay after processing 
	a BOSS object in order to allow other processes -- possibly with 
	lower priority -- to run. The method is executed inside of a process 
	that blocks whenever nothing is to be read from the socket."

	| delay |
	delay := Delay forMilliseconds: self class timeout.
	[socket successful] whileTrue: [
		self nextObject.
		delay wait]!

writeProcessLoop
	"Sit in this loop until the socket is closed. We delay after processing 
	a BOSS object in order to allow other processes -- possibly with 
	lower priority -- to run. This method runs inside of a process which 
	looks at messages placed on a shared queue. The next message 
	sent to the writeQueue object will cause this process to block 
	pending a write to the queue."

	| delay |
	delay := Delay forMilliseconds: self class timeout.
	[socket successful] whileTrue: [
		self nextPutDirectly: self writeQueue next.
		delay wait]! !

!TransportInterface methodsFor: 'private'!

nextObject
	"When an object can be obtained from the BOSS-based interface, 
	invoke the change dependency mechanism. If a nil object is 
	read, close the socket."

	| anObject |
	anObject := self next.
	anObject isNil
		ifTrue: [self close]
		ifFalse: [self changed: #socketIO with: anObject]! !

!TransportInterface methodsFor: 'specialized IO'!

next
	"Read a BOSS object off of the stream. If an error occurs, return 
	nil which will be interpreted in such a way as to close the socket 
	interface. As this method is been executed within a read process, 
	it ends up blocked within the readWait method invoked when 
	reading a BOSS object."

	socket errorReporter errorSignal handle: [:exception | self close. ^nil] do: [stream atEnd ifFalse: [^self bossRead next]].
	^nil!

nextPut: anObject 
	"Write a BOSS object to the writeQueue. The write process will then 
	write the object to the socket when it can."

	self writeQueue nextPut: anObject.
	Processor yield.!

nextPutDirectly: anObject 
	"Write a BOSS object to the stream. If an error occurs, close the socket. 
	A nil message ends up being seen by the other end which, in turn, 
	causes the other TransportInterface to close. Note that we could 
	block on the write which is possibly not what is required for a 
	server."

	(stream isKindOf: Stream)
		ifTrue: [socket errorReporter errorSignal
				handle: 
					[:exception | 
					self close.
					exception return]
				do: 
					[self bossWrite nextPut: anObject.
					stream commit]]! !

!TransportInterface methodsFor: 'private initialize-release'!

bossReadInit: aStream 
	"This initializes the boss read stream"

	^self bossRead: (BinaryObjectStorage onOldNoScan: aStream)!

bossWriteInit: aStream 
	"This initializes the boss write stream"

	^self bossWrite: (BinaryObjectStorage onNew: aStream)!

closeRead
	"Close down the read process"

	| aProcess |
	readProcess notNil
		ifTrue: 
			[aProcess := readProcess.
			readProcess == Processor activeProcess ifTrue: [[self close] fork].
			readProcess := nil.
			aProcess terminate]!

closeSocket
	"Close the socket interface"

	| tmp |
	((socket isKindOf: self class defaultConcreteClass)
		and: [socket successful])
		ifTrue: 
			[tmp := socket.
			socket := nil.
			tmp close]!

closeStream
	"Close the stream"

	| tmp |
	(stream isKindOf: Stream)
		ifTrue: 
			[tmp := stream.
			stream := nil.
			self changed: #closedSocket.
			tmp close]!

closeWrite
	"Close down the write process"

	| aProcess |
	writeProcess notNil
		ifTrue: 
			[aProcess := writeProcess.
			writeProcess == Processor activeProcess ifTrue: [[self close] fork].
			writeProcess := nil.
			aProcess terminate]!

setupInterface
	"This method sets up connection and stream correctly for the 
	socket-based interface."

	self setupInterfaceOn: self socket!

setupInterfaceOn: aSocket
	"This method sets up connection and stream correctly for the socket-based interface.
	We initialize the BOSS streams when reading and writing as we have to check/write
	header information at the start of the stream."

	self socket: aSocket.

	self connection: ExternalConnection new.
	self connection input: aSocket; output: aSocket.
	self stream: self connection readAppendStream.

	"Generate the processes associated with the interface. This allows
	for asynchronous processing of messages"

	self readProcess: ([self readProcessLoop] fork).
	self writeProcess: ([self writeProcessLoop] fork).

	"Now initialize the shared queue"
	self writeQueue: SharedQueue new.! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TransportInterface class
	instanceVariableNames: ''!


!TransportInterface class methodsFor: 'instance creation'!

newAtHost: aHost port: aPortNumber 
	"This creates a new transport interface using the requested port 
	number on a specific host"

	^self
		newAtHost: aHost
		port: aPortNumber
		handler: 
			[:exception | 
			Dialog warn: 'Server at port number ' , aPortNumber printString , ' not running on host ' , aHost , '.'.
			^nil]!

newAtHost: aHost port: aPortNumber handler: aBlock
	"This creates a new transport interface using the requested port 
	number on a specific host"

	self defaultConcreteClass errorReporter errorSignal
		handle: 
			[:exception | 
			aBlock value: exception.
			^nil]
		do: [^self newOn: (self defaultConcreteClass newTCPclientToHost: aHost port: aPortNumber)]!

newAtPort: aPortNumber 
	"This creates a new transport interface using the requested port 
	number"

	^self newAtPort: aPortNumber
		handler: 
			[:exception | 
			Dialog warn: 'Server at port number ' , aPortNumber printString , ' cannot be created.'.
			^nil]!

newAtPort: aPortNumber handler: aBlock
	"This creates a new transport interface using the requested port 
	number"

	| aSocket |
	self defaultConcreteClass errorReporter errorSignal
		handle: 
			[:exception | 
			aBlock value: exception.
			^nil]
		do: 
			[aSocket := self defaultConcreteClass newTCPserverAtPort: aPortNumber.
			aSocket listenFor: self queueSize.
			^(self new) socket: aSocket; yourself]!

newOn: aSocket 
	"This is used to create multiple clients to a given server. A read 
	process is spawned for each socket-based interface."

	^self new setupInterfaceOn: aSocket! !

!TransportInterface class methodsFor: 'class constants'!

defaultConcreteClass
	^SocketAccessor.!

queueSize
	"Size of queue for waiting connections"
	^1.!

timeout
	"The number of milliseconds that we're willing to wait for bytes to 
	arrive on socket"

	^100!

version
	^'Version 1.0a\ for VisualWorks 2.0\ Written by: Tony White' withCRs.! !

!TransportInterface class methodsFor: 'documentation'!

usage
	"A TransportInterface is used by both SmalltalkServer and 
	SmalltalkClient classes. A TransportInterface uses two processes; 
	one for reading and one for writing. The read process deals with 
	message objects sent by a remote Smalltalk and blocks waiting for 
	input on the socket. The write process blocks on a shared queue 
	which is written to by a separate process that wants to do output on 
	the socket. This latter process ensures that we never block on a 
	write -- only the write process blocks. 
	
	A TransportInterface is designed to support socket-based 
	communication between Smalltalk 80 images. The Binary Object 
	Storage System (BOSS) is used to transport arbitrarily complex 
	objects. A user need only send the message writeTerm: to an 
	instance of a TransportInterface with the message object as 
	argument. The message object is written to a shared queue which is 
	accessed by the write process forked when the interface is first set 
	up. In this way, the process doing the write will never block; only 
	the write process blocks. 
	
	Processing of message objects received over the socket interface can 
	be achieved using one of two alternatives. First, the application and 
	callback instance variables store the object and method to which the 
	received message object is dispatched. Alternatively, an object can 
	be made a dependent of the transport interface and have the 
	received message object dispatched to it via the changed: message. 
	The aspect used in #socketIO. It is possible to use both methods 
	simultaneously. Naturally, the dependency mechanism can be used 
	to achieve a broadcast mechanism. 
	
	An example of the former use is in the SmalltalkInterface 
	exampleClient: method. An example of the latter use can be found 
	in the SocketMsgReceiver update:with:from: method."

	^self error: 'Documentation only'! !

!TransportInterface class methodsFor: 'examples'!

exampleClient
	| aClient |
	aClient := self newAtHost: SocketAccessor getHostname port: 3603.
	aClient notNil
		ifTrue: 
			[aClient setupInterface;
				addDependent: SocketMsgReceiver new.
			aClient nextPut: 'Hello from the client'.
			(Delay forSeconds: 1) wait.
			aClient close]!

exampleServer
	| aClientInterface aServer |
	aServer := self newAtPort: 3603.
	aServer notNil
		ifTrue: 
			[aClientInterface := self newOn: aServer socket accept.
			aClientInterface addDependent: SocketMsgReceiver new.
			aClientInterface nextPut: 'Hello from server'.
			(Delay forSeconds: 1) wait.
			aClientInterface close.
			aServer close]! !

Model subclass: #SocketMsgReceiver
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smalltalk-Interface'!
SocketMsgReceiver comment:
'I am here as an example of a class which is set as a dependent of 
either SmalltalkServer or SmalltalkClient. I just write to the Transcript
whenever I receive an object over the interface.

Written by: Tony White (arpw@bnr.ca) and Dwight Deugo (ddeugo@bnr.ca)

This source code is available to all with no restrictions. We only ask that  
proper credit be passed on so that we might hear from those who have found  
the SocketMsgReceiver useful.
'!


!SocketMsgReceiver methodsFor: 'updating'!

update: aSymbol with: anObject from: aSender 
	"This processes socket messages that are received from 
	other Smalltalk images"

	aSymbol == #socketIO
		ifTrue: [Transcript show: self class name , ': ' , anObject printString; cr]
		ifFalse: [super
				update: aSymbol
				with: anObject
				from: aSender]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SocketMsgReceiver class
	instanceVariableNames: ''!


!SocketMsgReceiver class methodsFor: 'examples'!

exampleServer
	"SocketMsgReceiver exampleServer"

	| aServer |
	aServer := SmalltalkServer newAtPort: 3603.
	aServer notNil ifTrue: [aServer addDependent: SocketMsgReceiver new]! !

Model subclass: #SmalltalkInterface
	instanceVariableNames: 'application callback transport closeBlock '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smalltalk-Interface'!
SmalltalkInterface comment:
'The SmalltalkInterface class provides a means for communicating between two Smalltalk images that are running simultaneously.
BOSS streams are used to transport objects of arbitrary complexity. A client server relationship has to be established between
the two running images. Two mechanisms are provided for the dispatching of objects received via the interface. A callback mechanism
is supported and will be used if the application object responds to the method indicated in the callback instance variable. Otherwise,
a dependency mechanism can be used to dispatch objects to instances that express an interest in the interface.

This is an abstract class. The classes that do the real work are SmalltalkClient and SmalltalkServer.

Instance variable	Meaning
==============	=======
application			<any object class> An instance using the callback mechanism for transported-object dispatch.
callback				<Symbol> A Symbol indicating the instance method to be invoked with the object received.
transport			<TransportInterface> the transport object used for communications.
closeBlock			<BlockClosure> the block of code executed when the interface closes.

Written by: Tony White (arpw@bnr.ca) and Dwight Deugo (ddeugo@bnr.ca).

This source code is available to all with no restrictions. We only ask that  
proper credit be passed on so that we might hear from those who have found  
the SmalltalkInterface useful.


'!


!SmalltalkInterface methodsFor: 'accessing'!

application
	"This returns the instance variable"

	^application!

application: anObject 
	"This sets the instance variable"

	^application := anObject!

callback
	"This returns the instance variable"

	^callback!

callback: anObject 
	"This is provided in order to support the examples"

	^callback := anObject!

transport
	"This returns the instance variable"

	^transport!

transport: aTransportInterface 
	"This sets the instance variable"

	^transport := aTransportInterface! !

!SmalltalkInterface methodsFor: 'private'!

callbackMethod: anObject 
	"This is provided in order to support the examples"

	Transcript show: anObject printString; cr.! !

!SmalltalkInterface methodsFor: 'initialize-release'!

close
	"Close the interface"

	| tmp |
	tmp := self transport.
	tmp notNil
		ifTrue: 
			[self transport: nil.
			closeBlock value.
			tmp close]!

initialize
	closeBlock := [self].!

onCloseDo: aBlock 
	"This sets the instance variable. When the underlying transport 
	interface closes, this block of code is evaluated"

	^closeBlock := aBlock! !

!SmalltalkInterface methodsFor: 'updating'!

update: aSymbol with: anObject from: aSender 
	"This processes socket messages that are received from 
	other Smalltalk images"

	aSymbol == #socketIO
		ifTrue: 
			[(self application respondsTo: self callback)
				ifTrue: [self application perform: self callback with: anObject].
			dependents
				update: aSymbol
				with: anObject
				from: aSender]
		ifFalse: [super
				update: aSymbol
				with: anObject
				from: aSender]! !

!SmalltalkInterface methodsFor: 'specialized IO'!

nextOn: aTransportInterface 
	"This obtains an object from a specific client/server. Note 
	that this method could block the process invoking it."

	^aTransportInterface next!

nextPut: anObject on: aTransportInterface 
	"This sends an object to a specific client/server"

	aTransportInterface nextPut: anObject! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SmalltalkInterface class
	instanceVariableNames: ''!


!SmalltalkInterface class methodsFor: 'examples'!

exampleClientAtPort: aPort 
	"Try communicating locally"
	"[SmalltalkInterface exampleClientAtPort: 3603] fork"

	self exampleClientAtPort: aPort host: SocketAccessor getHostname.!

exampleClientAtPort: aPort host: aHostName
	"[SmalltalkInterface exampleClientAtPort: 3603 host: 'bgtyha'] fork"

	| aClient |
	aClient := SmalltalkClient newAtHost: aHostName port: aPort.
	aClient notNil
		ifTrue: 
			[aClient application: aClient;
			callback: #callbackMethod:;
			onCloseDo: [Transcript show: 'SmalltalkClient terminating'; cr];
			nextPut: 'Hello from client';
			nextPut: (OrderedCollection
					with: ColorValue black
					with: 2.9
					with: (TransportInterface new)).
			aClient nextPut: 'I am done'.
			(Delay forSeconds: 5) wait.
			aClient close]!

exampleServerAtPort: aPort 
	"[SmalltalkInterface exampleServerAtPort: 3603] fork"

	| aServer |
	aServer := SmalltalkServer newAtPort: aPort.
	aServer notNil
		ifTrue: 
			[aServer application: aServer;
			callback: #callbackMethod:;
			onNewClientDo: [:aTransportInterface | Transcript show: 'SmalltalkServer has new client'; cr];
			onClientCloseDo: [:aTransportInterface | Transcript show: 'SmalltalkServer client died'; cr]]! !

!SmalltalkInterface class methodsFor: 'cleanup'!

release
	"SmalltalkServer release"
	"SmalltalkClient release"

	self allInstances do: [:aSmalltalkInterface | aSmalltalkInterface close].
	ObjectMemory globalGarbageCollect! !

!SmalltalkInterface class methodsFor: 'instance creation'!

new
	^ super new initialize.! !

SmalltalkInterface subclass: #SmalltalkClient
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smalltalk-Interface'!
SmalltalkClient comment:
'I represent the client side of a socket-based interface between two Smalltalk images.
I expect to be able to connect to a (possibly remote) Smalltalk server. 

Written by: Tony White (arpw@bnr.ca) and Dwight Deugo (ddeugo@bnr.ca)

This source code is available to all with no restrictions. We only ask that  
proper credit be passed on so that we might hear from those who have found  
the SmalltalkClient useful.
'!


!SmalltalkClient methodsFor: 'specialized IO'!

next
	"This obtains a object from the server in BOSS format. Note 
	that this method could block the process invoking it."

	^self nextOn: self transport!

nextPut: anObject 
	"This sends an object to the server in BOSS format"

	self nextPut: anObject on: self transport! !

!SmalltalkClient methodsFor: 'updating'!

update: aSymbol with: anObject from: aSender 
	"This deals with socket closure"

	aSymbol == #closedSocket
		ifTrue: [self close]
		ifFalse: [super
				update: aSymbol
				with: anObject
				from: aSender]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SmalltalkClient class
	instanceVariableNames: ''!


!SmalltalkClient class methodsFor: 'instance creation'!

newAtHost: aHost port: aPort 
	^self newAtHost: aHost port: aPort handler: [:exception | exception].!

newAtHost: aHost port: aPort handler: aBlock
	| anInterface aTransport |
	aTransport := TransportInterface newAtHost: aHost port: aPort handler: aBlock.
	^aTransport notNil
		ifTrue: 
			[anInterface := self new.
			(anInterface transport: aTransport) setupInterface.
			aTransport addDependent: anInterface]
		ifFalse: [nil]! !

SmalltalkInterface subclass: #SmalltalkServer
	instanceVariableNames: 'listenerProcess interfaces clientCloseBlock newClientBlock '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smalltalk-Interface'!
SmalltalkServer comment:
'I represent the server side of a socket-based interface between two Smalltalk images.

Instance variable		Meaning
==============		=======
interfaces				<OrderedCollection> a collection of TransportInterface objects.
listenerProcess			<Process> a Process that sits and listens for new connections.
clientCloseBlock		<BlockClosure> invoked whenever a client connection closes.
newClientBlock			<BlockClosure> invoked whenever a new client connection is created. 

Written by: Tony White (arpw@bnr.ca) and Dwight Deugo (ddeugo@bnr.ca)

This source code is available to all with no restrictions. We only ask that  
proper credit be passed on so that we might hear from those who have found  
the SmalltalkServer useful.
'!


!SmalltalkServer methodsFor: 'private'!

closeInterfaces
	"This method closes all interfaces to clients"

	(interfaces respondsTo: #do:)
		ifTrue: [interfaces copy do: [:aTransportInterface | aTransportInterface close]].
	interfaces := nil!

stopListening
	"This ensures that we don't listen for any more connections"

	| aProcess |
	listenerProcess notNil
		ifTrue: 
			[aProcess := listenerProcess.
			listenerProcess := nil.
			aProcess terminate]! !

!SmalltalkServer methodsFor: 'initialize-release'!

close
	"Close all sockets that are currently open"

	super close.
	self stopListening.
	self closeInterfaces!

initialize
	super initialize.
	clientCloseBlock := newClientBlock := [:obj | obj].!

onClientCloseDo: aBlock 
	"This sets the instance variable"

	^clientCloseBlock := aBlock!

onNewClientDo: aBlock 
	"This sets the instance variable"

	^newClientBlock := aBlock! !

!SmalltalkServer methodsFor: 'accessing'!

interfaces
	"This returns the instance variable"

	^interfaces!

interfaces: anOrderedCollection 
	"This sets the instance variable"

	^interfaces := anOrderedCollection!

listenerProcess
	"This returns the instance variable"

	^listenerProcess!

listenerProcess: aProcess 
	"This sets the instance variable"

	^listenerProcess := aProcess! !

!SmalltalkServer methodsFor: 'client access'!

addClient: aTransportInterface 
	"This adds a new client to our list of clients. We make ourself 
	a dependent of the interface in order to catch close messages"

	self interfaces add: aTransportInterface.
	aTransportInterface addDependent: self.
	newClientBlock value: aTransportInterface!

removeClient: aTransportInterface 
	"This removes a client from our list of clients"

	self interfaces remove: aTransportInterface ifAbsent: ["Just don't care"].
	aTransportInterface removeDependent: self.
	clientCloseBlock value: aTransportInterface! !

!SmalltalkServer methodsFor: 'message processing'!

listenerProcessLoop: aPortNumber handler: aBlock
	"Sit in this loop waiting for new connections. I can be a REAL server 
	this way. As this method is executing within a process spawned 
	when the SmalltalkServer instance was created, it is blocked waiting 
	for an accept. So, all that happens is I do an accept -- creating a 
	new client -- and then wait for another accept."

	| childSocket newClient |
	transport := TransportInterface newAtPort: aPortNumber handler: aBlock.
	[transport notNil]
		whileTrue: 
			[childSocket := transport socket accept.
			childSocket isNil
				ifFalse: 
					["Now we have another client"
					newClient := TransportInterface newOn: childSocket.
					self addClient: newClient]]! !

!SmalltalkServer methodsFor: 'updating'!

update: aSymbol with: anObject from: aSender 
	"This deals with socket closure"

	aSymbol == #closedSocket
		ifTrue: [self removeClient: aSender]
		ifFalse: [super
				update: aSymbol
				with: anObject
				from: aSender]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SmalltalkServer class
	instanceVariableNames: ''!


!SmalltalkServer class methodsFor: 'instance creation'!

newAtPort: aPortNumber
	"Create a new server at a given port"
	^self newAtPort: aPortNumber handler: [:exception | exception].!

newAtPort: aPortNumber handler: aBlock
	"Create a new server at a given port"
	| aServer |
	aServer := self new.
	^aServer interfaces: OrderedCollection new; 
			listenerProcess: ([aServer listenerProcessLoop: aPortNumber handler: aBlock] fork); 
			yourself.! !

